"""
AutoUpdateServer: Automatically watches directory changes and updates connected clients
"""
import os
import time
import json
import logging
from pathlib import Path
from http.server import HTTPServer, BaseHTTPRequestHandler
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from typing import Dict, Optional, Set
import threading
from .utils import calculate_checksum

class ChangeHandler(FileSystemEventHandler):
    """Handle file system changes and update manifest"""
    def __init__(self, source_dir: str, manifest_path: str):
        self.source_dir = Path(source_dir)
        self.manifest_path = Path(manifest_path)
        self.pending_changes: Set[Path] = set()
        self.lock = threading.Lock()
        self._initialize_manifest()
        
    def _initialize_manifest(self):
        """Create initial manifest if it doesn't exist"""
        if not self.manifest_path.exists():
            manifest = {
                "version": "1.0.0",
                "last_update": time.time(),
                "files": {}
            }
            self._save_manifest(manifest)
    
    def _save_manifest(self, manifest: Dict):
        """Save manifest atomically"""
        temp_path = self.manifest_path.with_suffix('.tmp')
        try:
            with open(temp_path, 'w') as f:
                json.dump(manifest, f, indent=2)
            os.replace(str(temp_path), str(self.manifest_path))
        except Exception as e:
            logging.error(f"Failed to save manifest: {e}")
            if temp_path.exists():
                temp_path.unlink()

    def _update_manifest(self):
        """Update manifest with pending changes"""
        try:
            with self.lock:
                if not self.pending_changes:
                    return
                
                with open(self.manifest_path) as f:
                    manifest = json.load(f)
                
                manifest["last_update"] = time.time()
                
                for file_path in self.pending_changes:
                    rel_path = str(file_path.relative_to(self.source_dir))
                    if file_path.exists():
                        manifest["files"][rel_path] = {
                            "checksum": calculate_checksum(file_path),
                            "modified": os.path.getmtime(file_path)
                        }
                    else:
                        manifest["files"].pop(rel_path, None)
                
                self._save_manifest(manifest)
                self.pending_changes.clear()
                
        except Exception as e:
            logging.error(f"Failed to update manifest: {e}")

    def on_modified(self, event):
        if event.is_directory:
            return
        with self.lock:
            self.pending_changes.add(Path(event.src_path))
        self._update_manifest()

    def on_created(self, event):
        if event.is_directory:
            return
        with self.lock:
            self.pending_changes.add(Path(event.src_path))
        self._update_manifest()

    def on_deleted(self, event):
        if event.is_directory:
            return
        with self.lock:
            self.pending_changes.add(Path(event.src_path))
        self._update_manifest()

class UpdateRequestHandler(BaseHTTPRequestHandler):
    """Handle update requests from clients"""
    def do_GET(self):
        try:
            if self.path == '/manifest':
                self._send_manifest()
            elif self.path.startswith('/download/'):
                self._send_file(self.path[9:])
            else:
                self.send_error(404)
        except Exception as e:
            logging.error(f"Error handling request: {e}")
            self.send_error(500)

    def _send_manifest(self):
        """Send current manifest to client"""
        try:
            manifest_path = Path(self.server.source_dir) / '.manifest.json'
            if not manifest_path.exists():
                self.send_error(404)
                return
                
            with open(manifest_path) as f:
                manifest = json.load(f)
            
            self.send_response(200)
            self.send_header('Content-Type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps(manifest).encode())
            
        except Exception as e:
            logging.error(f"Error sending manifest: {e}")
            self.send_error(500)

    def _send_file(self, file_path: str):
        """Send requested file to client"""
        try:
            full_path = Path(self.server.source_dir) / file_path
            if not full_path.exists() or not full_path.is_file():
                self.send_error(404)
                return
            
            self.send_response(200)
            self.send_header('Content-Type', 'application/octet-stream')
            self.end_headers()
            
            with open(full_path, 'rb') as f:
                self.wfile.write(f.read())
                
        except Exception as e:
            logging.error(f"Error sending file: {e}")
            self.send_error(500)

class AutoUpdateServer:
    """Server that automatically updates connected clients when files change"""
    def __init__(self, source_dir: str, port: int = 8000):
        self.source_dir = Path(source_dir)
        self.port = port
        self.manifest_path = self.source_dir / '.manifest.json'
        
        # Initialize components
        self.observer = Observer()
        self.handler = ChangeHandler(str(self.source_dir), str(self.manifest_path))
        self.server = None
        self.server_thread = None
        
    def start(self):
        """Start the update server and file watcher"""
        try:
            # Start file system observer
            self.observer.schedule(self.handler, str(self.source_dir), recursive=True)
            self.observer.start()
            
            # Start HTTP server
            class UpdateServer(HTTPServer):
                source_dir = str(self.source_dir)
            
            self.server = UpdateServer(('0.0.0.0', self.port), UpdateRequestHandler)
            self.server_thread = threading.Thread(target=self.server.serve_forever)
            self.server_thread.daemon = True
            self.server_thread.start()
            
            logging.info(f"Auto-update server started on port {self.port}")
            logging.info(f"Watching directory: {self.source_dir}")
            
        except Exception as e:
            logging.error(f"Failed to start server: {e}")
            self.stop()
            raise
    
    def stop(self):
        """Stop the update server and file watcher"""
        if self.server:
            self.server.shutdown()
            self.server.server_close()
        if self.observer:
            self.observer.stop()
            self.observer.join()
        if self.server_thread:
            self.server_thread.join()
        logging.info("Auto-update server stopped")

def start_auto_update_server(directory: str, port: int = 8000) -> AutoUpdateServer:
    """Create and start an auto-update server"""
    server = AutoUpdateServer(directory, port)
    server.start()
    return server
