"""
Core functionality for RedXSecureUpdater with optimized performance
"""
import os
import json
import logging
import subprocess
import threading
import time
import requests
import ssl
import socket
from hashlib import sha256
import hmac
from http.server import HTTPServer, BaseHTTPRequestHandler
from pathlib import Path
from typing import Optional, List, Dict, Tuple
from datetime import datetime
import shutil

from .monitor import FileMonitor
from .version_control import VersionControl

class UpdateRequestHandler(BaseHTTPRequestHandler):
    """Handle secure update requests with authentication"""
    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger("RedXSecureUpdater.UpdateRequestHandler")
        super().__init__(*args, **kwargs)

    def do_GET(self):
        try:
            # Enhanced security check
            if not self._verify_request():
                self.logger.warning(f"Access denied from {self.client_address[0]}")
                self.send_error(403, "Access denied")
                return

            if self.path == '/check_update':
                self._send_update_status()
            elif self.path == '/download_update':
                self._send_update_package()
            else:
                self.send_error(404)
        except Exception as e:
            self.logger.error(f"Update request error: {e}")
            self.send_error(500)

    def _verify_request(self) -> bool:
        """Enhanced request verification with token and origin checks"""
        try:
            token = self.headers.get('X-Update-Token')
            if not token:
                return False

            # Generate token using shared secret
            secret = b"secure-update"
            message = b"update-request"
            expected = hmac.new(secret, message, sha256).hexdigest()

            # Use constant-time comparison to prevent timing attacks
            if not hmac.compare_digest(token, expected):
                return False

            return True

        except Exception as e:
            self.logger.error(f"Token verification failed: {e}")
            return False

    def _send_update_status(self):
        """Send current update status with enhanced security"""
        try:
            updater = self.server.updater
            if not updater:
                self.send_error(500, "Server configuration error")
                return

            status = {
                'version': updater.version_control.get_version() if updater.version_control else "1.0.0",
                'update_available': updater.has_pending_updates(),
                'checksum': updater.get_update_checksum()
            }
            self._send_json(status)
        except Exception as e:
            self.logger.error(f"Error sending update status: {e}")
            self.send_error(500)

    def _send_update_package(self):
        """Send update package with enhanced security checks"""
        try:
            updater = self.server.updater
            if not updater:
                self.send_error(500, "Server configuration error")
                return

            update_file = updater.create_update_package()
            if not update_file:
                self.send_error(404, "No updates available")
                return

            self.send_response(200)
            self.send_header('Content-Type', 'application/zip')
            self.send_header('X-Update-Checksum', updater.get_update_checksum())
            self.send_header('X-Content-Security-Policy', "default-src 'self'")
            self.end_headers()

            with open(update_file, 'rb') as f:
                self.wfile.write(f.read())

        except Exception as e:
            self.logger.error(f"Failed to send update package: {e}")
            self.send_error(500)

    def _send_json(self, data):
        """Send JSON response"""
        try:
            self.send_response(200)
            self.send_header('Content-Type', 'application/json')
            self.end_headers()
            self.wfile.write(json.dumps(data).encode())
        except Exception as e:
            self.logger.error(f"Error sending JSON response: {e}")
            self.send_error(500)

class UpdateServer(HTTPServer):
    """Secure update server with SSL support"""
    def __init__(self, updater, server_address):
        self.updater = updater
        super().__init__(server_address, UpdateRequestHandler)
        self.allow_reuse_address = True

class RedXSecureUpdater:
    """Enhanced secure update manager with built-in server"""
    def __init__(
        self,
        directory: str,
        version: Optional[str] = None,
        ssl_cert: Optional[str] = None,
        ssl_key: Optional[str] = None,
        check_interval: int = 5
    ):
        self.directory = Path(directory).resolve()
        self._setup_logging()
        self._init_version_control(version)

        # Create necessary directories
        os.makedirs(self.directory / '.redx', exist_ok=True)

        # Initialize monitor first
        self.is_running = True
        self.monitor = None
        self._init_monitor()

        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key
        self.server = None
        self.server_port = None
        self.server_thread = None
        self.check_interval = max(5, check_interval)

        # Wait for monitor to be ready
        if self.monitor and self.monitor.wait_ready():
            self._start_monitor_thread()
            # Force initial update check after monitor is ready
            if self.has_pending_updates():
                try:
                    self._ensure_server_running()
                    self.logger.info("Server started automatically due to pending updates")
                except Exception as e:
                    self.logger.error(f"Failed to start server automatically: {e}")
        else:
            self.logger.error("Failed to initialize monitor")

    def _setup_logging(self):
        """Setup logging with performance-friendly configuration"""
        self.logger = logging.getLogger("RedXSecureUpdater")
        if not self.logger.handlers:
            handler = logging.StreamHandler()
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)

    def _init_version_control(self, version: Optional[str] = None):
        """Initialize version control"""
        self.version_control = VersionControl(str(self.directory))
        if version:
            self.version_control.set_version(version)

    def _init_monitor(self):
        """Initialize file monitor"""
        self.monitor = FileMonitor(str(self.directory))
        self.monitor.start()

    def has_pending_updates(self) -> bool:
        """Check for pending updates with caching"""
        try:
            if not hasattr(self, '_last_check_time'):
                self._last_check_time = 0
                self._last_check_result = False

            current_time = time.time()
            if current_time - self._last_check_time < self.check_interval:
                return self._last_check_result

            if self.monitor:
                changes = self.monitor.get_changes()
                self._last_check_result = bool(changes)
                self._last_check_time = current_time
                self.logger.debug(f"Update check: {self._last_check_result}")
                return self._last_check_result
            return False

        except Exception as e:
            self.logger.error(f"Error checking updates: {e}")
            return False

    def _ensure_server_running(self):
        """Ensure update server is running with SSL if configured"""
        if self.server is not None:
            return

        try:
            port = self._find_free_port()
            server = UpdateServer(self, ('0.0.0.0', port))

            if self.ssl_cert and self.ssl_key:
                try:
                    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
                    context.load_cert_chain(self.ssl_cert, self.ssl_key)
                    server.socket = context.wrap_socket(server.socket, server_side=True)
                except Exception as e:
                    self.logger.error(f"SSL configuration failed: {e}")
                    raise

            self.server = server
            self.server_port = port

            self.server_thread = threading.Thread(target=server.serve_forever)
            self.server_thread.daemon = True
            self.server_thread.start()

            self.logger.info(f"Update server started on port {port}")

        except Exception as e:
            self.logger.error(f"Failed to start server: {e}")
            if hasattr(self, 'server') and self.server:
                try:
                    self.server.shutdown()
                    self.server.server_close()
                except Exception:
                    pass
                self.server = None
            raise

    def _find_free_port(self, start_port=8443) -> int:
        """Find available port for server"""
        port = start_port
        while port < start_port + 100:
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.bind(('', port))
                    return port
            except OSError:
                port += 1
        raise RuntimeError("No free ports found")

    def get_update_checksum(self) -> str:
        """Calculate checksum for pending updates"""
        if not self.monitor:
            return ""

        changes = self.monitor.get_changes()
        if not changes:
            return ""

        hasher = sha256()
        for file_path, _ in changes:
            full_path = self.directory / file_path
            if full_path.exists():
                with open(full_path, 'rb') as f:
                    hasher.update(f.read())
        return hasher.hexdigest()

    def create_update_package(self) -> Optional[str]:
        """Create secure update package"""
        if not self.monitor:
            return None

        changes = self.monitor.get_changes()
        if not changes:
            return None

        try:
            import zipfile
            package_path = self.directory / '.redx' / 'update.zip'
            os.makedirs(os.path.dirname(package_path), exist_ok=True)

            with zipfile.ZipFile(package_path, 'w', zipfile.ZIP_DEFLATED) as zf:
                for file_path, _ in changes:
                    full_path = self.directory / file_path
                    if full_path.exists():
                        zf.write(full_path, arcname=file_path)

            return str(package_path)

        except Exception as e:
            self.logger.error(f"Failed to create update package: {e}")
            return None

    def stop(self):
        """Stop update server and monitoring"""
        self.is_running = False
        if hasattr(self, 'server') and self.server:
            try:
                self.server.shutdown()
                self.server.server_close()
            except Exception:
                pass
            self.server = None
        if self.monitor:
            try:
                self.monitor.stop()
            except Exception:
                pass
            self.monitor = None
        self.logger.info("Update server stopped")

    def schedule_update(self, schedule_time: str):
        """Schedule an update for a specified time"""
        self.logger.info(f"Update scheduled for {schedule_time}")

    def _start_monitor_thread(self):
        """Start background monitoring thread"""
        def monitor_loop():
            while self.is_running:
                try:
                    if self.has_pending_updates():
                        self._ensure_server_running()
                    time.sleep(self.check_interval)
                except Exception as e:
                    self.logger.error(f"Monitor error: {e}")

        self.monitor_thread = threading.Thread(target=monitor_loop, daemon=True)
        self.monitor_thread.start()


def RedXSecureUpdate(
    directory: str,
    version: Optional[str] = None,
    schedule_time: Optional[str] = None,
    app_command: Optional[str] = None,
    ssl_cert: Optional[str] = None,
    ssl_key: Optional[str] = None,
    message: Optional[str] = None
) -> bool:
    """Convenient update function with improved error handling"""
    try:
        if not directory:
            raise ValueError("Directory parameter is required")

        directory = str(Path(directory).resolve())
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        # Ensure .redx directory exists
        redx_dir = os.path.join(directory, '.redx')
        os.makedirs(redx_dir, exist_ok=True)

        updater = RedXSecureUpdater(
            directory=directory,
            version=version,
            ssl_cert=ssl_cert,
            ssl_key=ssl_key
        )

        # Handle scheduled updates
        if schedule_time:
            try:
                schedule_time = datetime.strptime(schedule_time, "%I:%M %p").strftime("%H:%M")
                updater.schedule_update(schedule_time)
                return True
            except ValueError as e:
                logging.error(f"Invalid schedule time format: {e}")
                return False

        # Handle immediate updates
        try:
            has_updates = updater.has_pending_updates()
            if version:
                # Force version update even if no files changed
                updater.version_control.set_version(version)
                has_updates = True

            if has_updates and app_command:
                try:
                    # Restart application if command provided
                    subprocess.Popen(app_command, shell=True)
                except Exception as e:
                    logging.error(f"Failed to execute app command: {e}")

            if message:
                updater.logger.info(f"Update Message: {message}")

            return True  # Return True if version was set or updates were found

        except Exception as e:
            logging.error(f"Update process failed: {e}")
            return False

        finally:
            try:
                updater.stop()
            except Exception:
                pass

    except Exception as e:
        logging.error(f"Update initialization failed: {e}")
        return False